/*
 *
 * Copyright (C) 2006 Mekensleep
 *
 *	Mekensleep
 *	24 rue vieille du temple
 *	75004 Paris
 *       licensing@mekensleep.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *  Igor Kravtchenko <igor@tsarevitch.org>
 *
 */

#ifndef UNDERWARE_VSERIAL_SCENESERIALIZER_H
#define UNDERWARE_VSERIAL_SCENESERIALIZER_H

#ifndef UNDERWARE_VSERIAL_USE_PCH
#include <map>
#include <vector>

#include <vserial/vserial.h>
#endif

ENTER_NAMESPACE_UNDERWARE

class DataIn;
class DataOut;
class Motion;
class Scene;
class SceneItem;
class SceneMesh;
class SceneNullObject;
class SceneLight;

class SceneSerializer {
public:

	class SaveOptions {
	public:
		bool bDontSaveAnimations;
	};

	UW_VSERIAL_API static bool load(const char *fileName, const char *basePath, Scene **res = NULL);
	UW_VSERIAL_API static bool load(DataIn &data, const char *basePath, Scene **res = NULL);

	UW_VSERIAL_API static bool save(Scene &scene, const char *fileName, SaveOptions *options = NULL);
	UW_VSERIAL_API static bool save(Scene &scene, DataOut &data, SaveOptions *options = NULL);

protected:
	bool load();
	void save();

	DataIn *in_;
	DataOut *out_;
	Scene *scene_;
	const char *path_;
	SaveOptions *saveOptions_;
	std::map<SceneItem*, int> items2id_;
	std::map<int, SceneItem*> ids2item_;
	std::map<SceneItem*, int> items2parent_;
	std::map<Motion*, bool> motions_;
	std::map<std::string, SceneItem*> mots2item_;

	// read functions
	bool readMESHchunk(SceneMesh &, int chunkSize);
	bool readNULOchunk(SceneNullObject &, int chunkSize);
	bool readLGTchunk(SceneLight &, int chunkSize);
	bool readCommonItemAttributes(int tagid, int chunkSize, SceneItem &);

	// write functions
	void writeNAMEchunk(const char *);
	void writeMESHchunk(const SceneMesh &);
	void writeLGTchunk(const SceneLight &);
	void writeNULOchunk(const SceneNullObject &);
	void writeCommonItemAttributes(const SceneItem &);
};

LEAVE_NAMESPACE

#endif // UNDERWARE_VSERIAL_SCENESERIALIZER_H
