/*
 *
 * Copyright (C) 2006 Mekensleep
 *
 *	Mekensleep
 *	24 rue vieille du temple
 *	75004 Paris
 *       licensing@mekensleep.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *  Igor Kravtchenko <igor@tsarevitch.org>
 *
 */

#include <vserial/stdafx.h>

#ifndef UNDERWARE_VSERIAL_USE_PCH
#include <vserial/vserial.h>
#include <vserial/skinning.h>
#endif

ENTER_NAMESPACE_UNDERWARE

Skinning::Skinning(const std::string &_name) :
name_(_name)
{
}

Skinning::~Skinning()
{
}

int Skinning::getIndexOfBone(SceneBone *_bone) const
{
	std::map<SceneBone*, FloatMap*>::const_iterator it = bones2FloatMap_.begin();
	int i = 0;
	for ( ; it != bones2FloatMap_.end(); ++it) {
		const SceneBone *bone = it->first;
		if (bone == _bone)
			return i;
		i++;
	}
	return -1;
}

bool Skinning::hasBone(SceneBone *_bone) const
{
	return getIndexOfBone(_bone) == -1 ? false : true;
}

void Skinning::setBone(SceneBone *_bone, FloatMap *_fmap)
{
	bones2FloatMap_[_bone] = _fmap;
}

std::vector<SceneBone*> Skinning::getBonesList() const
{
	std::vector<SceneBone*> res;
	std::map<SceneBone*, FloatMap*>::const_iterator it = bones2FloatMap_.begin();
	for ( ; it != bones2FloatMap_.end(); ++it) {
		SceneBone *bone = it->first;
		res.push_back(bone);
	}
	return res;
}

FloatMap* Skinning::getWeightMapByBone(SceneBone *_bone)
{
	std::map<SceneBone*, FloatMap*>::iterator it = bones2FloatMap_.find(_bone);
	if (it == bones2FloatMap_.end())
		return NULL;
	return it->second;
}

void Skinning::removeBone(SceneBone *_bone)
{
	std::map<SceneBone*, FloatMap*>::iterator it = bones2FloatMap_.begin();
	for ( ; it != bones2FloatMap_.end(); ++it) {
		SceneBone *bone = it->first;
		if (bone == _bone) {
			bones2FloatMap_.erase(it);
			return;
		}
	}
}

LEAVE_NAMESPACE
